#include <stdbool.h>
#include <stdint.h>

#include "disp.h"
#include "clock.h"

/* Internal clock time */
static struct {
	uint8_t seconds;
	uint8_t minutes;
	uint8_t hours;
} clock_time;

/* State machine for time setup */
static enum {
	CLOCK_RUNNING,
	CLOCK_SETUP_POS0,
	CLOCK_SETUP_POS1,
	CLOCK_SETUP_POS2,
	CLOCK_SETUP_POS3,
} clock_mode;

/* Convert internal time to display output */
static void clock_update(void)
{
	uint16_t time = 0;

	time |= (clock_time.hours   / 10) << 12;
	time |= (clock_time.hours   % 10) <<  8;
	time |= (clock_time.minutes / 10) <<  4;
	time |= (clock_time.minutes % 10) <<  0;
	disp_set_bcd(time);
}

bool clock_on_key(enum key_code code)
{
	bool consumed = false;
	uint8_t value;
	
	switch (code) {
	case KEY_TIME:
		/* Start setup mode */
		clock_mode = CLOCK_SETUP_POS3;
		disp_set_blnk_mask(1 << 3);
		clock_time.seconds = 0;
		consumed = true;
		break;
	case KEY_OK:
		/* Use new setting (after validation) */
		if (clock_time.hours > 23)
			clock_time.hours = 23;
		if (clock_time.minutes > 59)
			clock_time.minutes = 59;
		clock_mode = CLOCK_RUNNING;
		disp_set_blnk_mask(0);
		consumed = true;
		break;
	case KEY_DIG0:
	case KEY_DIG1:
	case KEY_DIG2:
	case KEY_DIG3:
	case KEY_DIG4:
	case KEY_DIG5:
	case KEY_DIG6:
	case KEY_DIG7:
	case KEY_DIG8:
	case KEY_DIG9:
		/* Implicit knowledge: they key codes for digits 0 to 9
		 * are follow each other without gaps.
		 */
		value = code - KEY_DIG0;
		/* Check which digit needs to be set. This advances also the
		 * state machine to the next digit and adjusts the blinking mask.
		 */
		switch (clock_mode) {
		case CLOCK_SETUP_POS0:
			clock_time.minutes = clock_time.minutes / 10 * 10;
			clock_time.minutes += value;
			clock_mode = CLOCK_SETUP_POS3;
			disp_set_blnk_mask(1 << 3);
			break;
		case CLOCK_SETUP_POS1:
			if (value < 6) {
				clock_time.minutes = clock_time.minutes % 10;
				clock_time.minutes += value * 10;
				clock_mode = CLOCK_SETUP_POS0;
				disp_set_blnk_mask(1 << 0);
			}
			break;
		case CLOCK_SETUP_POS2:
			clock_time.hours = clock_time.hours / 10 * 10;
			clock_time.hours += value;
			clock_mode = CLOCK_SETUP_POS1;
			disp_set_blnk_mask(1 << 1);
			break;
		case CLOCK_SETUP_POS3:
			if (value < 3) {
				clock_time.hours = clock_time.hours % 10;
				clock_time.hours += value * 10;
				clock_mode = CLOCK_SETUP_POS2;
				disp_set_blnk_mask(1 << 2);
			}
			break;
		}
		consumed = true;
		break;
	}
	
	/* If the key was used, immediately update display */
	if (consumed)
		clock_update();
	
	return consumed;
}

void clock_tick(void)
{
	/* Advance internal time only if in running mode, not during setup */
	if (clock_mode == CLOCK_RUNNING) {
		clock_time.seconds++;
		if (clock_time.seconds > 59) {
			clock_time.seconds = 0;
			clock_time.minutes++;
		}
		if (clock_time.minutes > 59) {
			clock_time.minutes = 0;
			clock_time.hours++;
		}
		if (clock_time.hours > 23) {
			clock_time.hours = 0;
		}
		
		clock_update();
	}
}

void clock_init(void)
{
	clock_time.seconds = 0;
	clock_time.minutes = 0;
	clock_time.hours = 0;
	clock_mode = CLOCK_RUNNING;
}
