#include <avr/io.h>
#include <avr/interrupt.h>

#include <stdint.h>
#include <stdbool.h>

#include "irq.h"

#include "rc5.h"
#include "main.h"
#include "disp.h"
#include "clock.h"

#include "timer.h"

typedef void (*timer_proc_t)(void);

struct timer_entry {
	const timer_proc_t   proc;     // timer callback procedure
	const uint16_t       reload;   // reload value for counter
	uint16_t             counter;  // tick counter
	bool                 ready;    // flag if ready to run
};

volatile struct timer_entry timer_table[] = {
	/* function         ticks */
	{  disp_mpx_tick,      2  },
	{  main_10ms,         10  },
	{  disp_blnk_tick,   250  },
	{  clock_tick,      1000  },
};
#define TIMER_NUM_ENTRIES  (sizeof(timer_table)/sizeof(timer_table[0]))

/* Runs every 100 microseconds from ISR context */
inline void timer_100us_isr(void)
{
	rc5_sample();
}

/* Runs every millisecond from ISR context */
inline void timer_1ms_isr(void)
{
	volatile struct timer_entry *entry = timer_table;
	
	while (entry != &timer_table[TIMER_NUM_ENTRIES]) {
		if (entry->counter > 0)
			entry->counter--;
		if (entry->counter == 0) {
			entry->counter = entry->reload;
			entry->ready = true;
		}
		entry++;
	}
}

///////////////////////////////////////////////////////////////////////

ISR(TIMER0_OVF_vect)
{
	static uint8_t ms_div;
	
	/* every 100 us */
	TCNT0 = (256 - 100);

	/* Call 100us first to reduce jitter */
	timer_100us_isr();
	
	/* Decrement divider for 1ms base */
	if (ms_div > 0)
		ms_div--;
	if (ms_div == 0) {
		/* reload with 1ms / 100us = 10 */
		ms_div = 10;
		timer_1ms_isr();
	}
}

void timer_dispatch(void)
{
	volatile struct timer_entry *entry = timer_table;

	while (entry != &timer_table[TIMER_NUM_ENTRIES]) {
		uint8_t lock;
		bool ready;
		
		irq_lock(&lock);
		ready = entry->ready;
		entry->ready = false;
		irq_unlock(lock);
		
		if (ready)
			entry->proc();
		entry++;
	}	
}

void timer_init(void)
{
	/* prescaler 1/8 */
	TCCR0 = (1 << CS01);
	/* enable interrupts */
	TIMSK |= (1 << TOIE0);
}
